=begin

  # Copyright 2011+, Jrg Bergmann

  # Permission to use this software for any purpose and without fee is hereby granted
  # Distribution of this software for commercial purpose is subject to:
  #  - the expressed, written consent of the author
  #  - the inclusion of the present copyright notice in all copies.

  # THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
  # IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  
  #----------------------------------------------------------------------------
  # Name        : dxf_export.rb
  # Version     : 0.0.3
  # Date        : 2011-10-22
  # Type			  : Ruby Script
  # Description	: Exports polygons|polylines|cirles|arcs|lines to .DXF
  # Usage			  : Just try ... ;)
  #               Tools->Export to DXF 
  #----------------------------------------------------------------------------
  
  # *** 2011-10-22 ***
  #   - improved objects assignment ( faster)
  #   - improved calculations (assuming that /2 returns int and not float as needed)
  #   - automatic unit detection
  
=end

require 'sketchup.rb'

module Dxf_export

  def Dxf_export.main(debug=nil)
  
    puts("\n--- DXF_export.rb debug active ----------------------------------") if debug
    
    unit_factor=get_unit_factor()
    if unit_factor == -1
      return nil if UI.messagebox("Unable to detect the actual unit setting ...\nGenerate Output in mm?", MB_YESNO) == 7
    end
    
    # get the active model
    model = Sketchup.active_model
   
    objects = Array.new # shapes of connected lines (edges only)
    shapes = Array.new # objects with the edges assigned to polygons|polylines|cirles|arcs|lines
    
    processed = Array.new # Collector for all processed edges
        
    # explode until there is no group left
    Dxf_export.explode_all
  
    # define the filename
    model_name = "Untitled" if (model_name = File.basename(model.path, ".skp")) == ""
    
    # get selected/all entities of the active model
    if model.selection.empty?
      return nil if UI.messagebox("Nothing selected!\nContinue with all entities?", MB_YESNO) == 7
        ents = model.entities
    else
      ents = model.selection
    end
    
    if ents.length == 0
      UI.messagebox("Drawing is empty!\nNothing to do!")
      return nil
    end
      
    if Sketchup.version[0,1].to_i >= 7
      model.start_operation("DXF_export", true)
    else
      model.start_operation "DXF_export"
    end
   
    Sketchup.status_text="Assigning objects ..."

    ents.each { |ent| 
      if not processed.index(ent)
        tmp = Array.new
        if ent.typename == "Edge"
          ent.all_connected.each { |e|
            tmp << e if e.typename == "Edge" # 'all_connected' returns faces too
            processed << e
          }
        end
        objects << tmp if tmp.length > 0
      end
    }

    Sketchup.status_text="Parsing objects ..."
    
    # parse all edges in each 'object'
    objects.each { |obj|
    
      lines = Array.new
      polylines = Array.new
      polygons = Array.new
      arcs = Array.new
      circles = Array.new
    
      obj.each { |ent|
        if ent.curve
          if ent.curve.typename != "ArcCurve"
            polylines << ent.curve if not polylines.index(ent.curve)
          elsif ent.curve.is_polygon?
            polygons << ent.curve if not polygons.index(ent.curve)
          else
            #if ent.curve.end_angle-ent.curve.start_angle==Math::PI*2
            if ent.curve.end_angle==Math::PI*2
              circles << ent.curve if not circles.index(ent.curve)
            else
              arcs << ent.curve if not arcs.index(ent.curve)
            end
          end
        else
            lines << (ent)
        end
      }
      shapes.push([polylines, polygons, circles, arcs, lines])
    }

    UI.messagebox("#{shapes.length} shapes found")
    
    Sketchup.set_status_text()
    
    save_name = UI.savepanel( "Save DXF file ...", File.dirname(model.path) , "#{model_name}.dxf" )

    model.commit_operation
    
    out_file = File.new( save_name , "w" )
    out_file.puts("0\nSECTION\n2\nENTITIES\n")
  
    id = 0
    shapes.each { |shape| 
      shape[0].each { |p| # polyline (currently polylines output to lines)
        p.each_edge { |l|
          out_file.puts("0\nLINE\n8\n#{id}\n10\n#{l.start.position[0].to_f*unit_factor}\n20\n#{l.start.position[1].to_f*unit_factor}\n30\n#{l.start.position[2].to_f*unit_factor}\n11\n#{l.end.position[0].to_f*unit_factor}\n21\n#{l.end.position[1].to_f*unit_factor}\n31\n#{l.end.position[2].to_f*unit_factor}\n")
        }
      }
      shape[1].each { |p| # polygone (currently polygone output to lines)
        id+=1
        p.each_edge { |l|
          out_file.puts("0\nLINE\n8\n#{id}\n10\n#{l.start.position[0].to_f*unit_factor}\n20\n#{l.start.position[1].to_f*unit_factor}\n30\n#{l.start.position[2].to_f*unit_factor}\n11\n#{l.end.position[0].to_f*unit_factor}\n21\n#{l.end.position[1].to_f*unit_factor}\n31\n#{l.end.position[2].to_f*unit_factor}\n")
        }
      }
      shape[2].each { |c| # circle
        out_file.puts("0\nCIRCLE\n8\n#{id}\n10\n#{c.center[0].to_f*unit_factor}\n20\n#{c.center[1].to_f*unit_factor}\n30\n#{c.center[2].to_f*unit_factor}\n40\n#{c.radius.to_f*unit_factor}\n210\n#{c.normal[0].to_f*unit_factor}\n220\n#{c.normal[1].to_f*unit_factor}\n230\n#{c.normal[2].to_f*unit_factor}\n")
      }
      shape[3].each { |a| # arc
        rad_to_deg=180.0/Math::PI
        
        # find orientation of arc
        vect=Geom::Vector3d.new(0.0,0.0,0.0)
        a.each_edge { |edg|
          vect[0]+=edg.start.position.x-a.center.x
          vect[1]+=edg.start.position.y-a.center.y
          vect[0]+=edg.end.position.x-a.center.x
          vect[1]+=edg.end.position.y-a.center.y
        }
        vect.length=-a.radius
        n=a.center+vect
        
        if a.center.y <= n.y
          angle=rad_to_deg*Geom::Vector3d.new(1.0,0.0,0.0).angle_between(vect)
        else
          angle=360.0-rad_to_deg*Geom::Vector3d.new(1.0,0.0,0.0).angle_between(vect)
        end
        
        rot=(360.0-rad_to_deg*a.end_angle)/2.0

        if debug
          Sketchup.active_model.entities.add_cpoint(a.center)
          Sketchup.active_model.entities.add_cline(a.center, n)
          Sketchup.active_model.entities.add_text(angle.to_s, n, Geom::Vector3d.new(-20.0,20.0,0.0))
        end
        
        out_file.puts("0\nARC\n8\n#{id}\n10\n#{a.center[0].to_f*unit_factor}\n20\n#{a.center[1].to_f*unit_factor}\n30\n#{a.center[2].to_f*unit_factor}\n40\n#{a.radius.to_f*unit_factor}\n50\n#{(angle+rot).to_f}\n51\n#{(angle-rot).to_f}\n210\n#{a.normal[0].to_f*unit_factor}\n220\n#{a.normal[1].to_f*unit_factor}\n230\n#{a.normal[2].to_f*unit_factor}\n")
      }
     shape[4].each { |l| # line
        out_file.puts("0\nLINE\n8\n#{id}\n10\n#{l.start.position[0].to_f*unit_factor}\n20\n#{l.start.position[1].to_f*unit_factor}\n30\n#{l.start.position[2].to_f*unit_factor}\n11\n#{l.end.position[0].to_f*unit_factor}\n21\n#{l.end.position[1].to_f*unit_factor}\n31\n#{l.end.position[2].to_f*unit_factor}\n")
      }
      id += 1
    }
    
      out_file.puts("0\nENDSEC\n0\nEOF\n")
      out_file.close

      Sketchup.set_status_text("#{save_name} saved ...") 

  end
  
  def Dxf_export.get_unit_factor
    # detect units
    str_unit = ""
    str_formatedLength = Sketchup.format_length(1)
    str_units=["\"", "'", "mm", "cm", "m"].each { |u|
      if not str_formatedLength.index(u) == nil
        str_unit = u
        break u
      end
    }
    
    case str_unit
      when "\""
        return 1
      when "'"
        return 1.0/12.0
      when "mm"
        return 25.4
      when "cm"
        return 2.54
      when "m"
        return 0.254
      else
        return -1
    end
  end
  
  def Dxf_export.explode_all
    Sketchup.set_status_text("Exploding all Groups and Components ...") 
    counter = 0
    Sketchup.active_model.entities.each { |e|
      if e.typename == "Group" || e.typename == "ComponentInstance"
        e.explode
        counter+=1
      end
    }
    # rerun if there were any otherwise all groups exploded
    counter > 0 ? Dxf_export.explode_all : Sketchup.set_status_text("Exploding done ...")
  end
  
  def Dxf_export.reload
    load("dxf_export.rb")
    puts("'dxf_export.rb' reloaded")
  end

end

if( not file_loaded?("dxf_export.rb") )
   add_separator_to_menu("Tools")
   # UI.menu("Tools").add_item("reload(Export to DXF)") { Dxf_export.reload }
   UI.menu("Tools").add_item("Export to DXF") { Dxf_export.main }
end

file_loaded("dxf_export.rb")
