# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   Recurve plugin for Google SketchUp
# Created by Diggory Blake on 22 April 2011
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

require 'sketchup.rb'

module Recurve
	def Recurve.can_select_curve?(ents)
		return false unless ents.length == 1
		ent = ents[0]
		
		return ent.is_a? Sketchup::Edge
	end
	
	def Recurve.can_select_curve_selection?()
		Sketchup.active_model and can_select_curve?(Sketchup.active_model.selection)
	end
	
	def Recurve.select_curve(ents, planar)
		return unless can_select_curve?(ents)
		
		edges = []
		
		points = []
		plane = nil
		
		findadjacent = lambda do |ent, fromvertex|
			unless edges.index(ent) then
				edges << ent
				ent.vertices.each do |vertex|
					if vertex.equal? fromvertex then
						next
					end
				
					if planar then
						if points.length < 3 then
							points << vertex.position
							if points.length == 3 then
								vec0 = points[1]-points[0]
								vec1 = points[2]-points[1]
								if vec0.parallel? vec1 then
									points.pop
								else
									plane = Geom.fit_plane_to_points(points)
								end
							end
						else
							unless vertex.position.on_plane? plane then
								break
							end
						end
					end
					
					vertex.edges.each do |edge|
						unless edge.soft? then
							findadjacent.call(edge, vertex)
						end
					end
				end
			end
		end
		
		findadjacent.call(ents[0], nil)
		
		Sketchup.active_model.selection.clear
		Sketchup.active_model.selection.add edges
		
		return edges
    end
	
	def Recurve.select_curve_selection(planar)
		select_curve Sketchup.active_model.selection, planar
	end

	def Recurve.can_recurve?(ents)
		return false unless ents.length > 1
		ents.each do |ent|
			return false unless ent.is_a? Sketchup::Edge
		end
		
		return true
	end
	
	def Recurve.recurve(ents)
		return unless can_recurve?(ents)
		
		Sketchup.active_model.start_operation "Recurve Edges"
		
		edges = []
		
		curvepoints = []
		
		direction = 0
		
		findadjacent = lambda do |ent, fromvertex|
			unless edges.index(ent) then
				edges << ent
				ent.vertices.each do |vertex|
					if direction == 0 then
						curvepoints.push(vertex.position)
					else
						curvepoints.unshift(vertex.position)
					end
				
					unless fromvertex then
						direction = 1
					end
				
					if vertex.equal? fromvertex then
						next
					end
					
					vertex.edges.each do |edge|
						if !edge.soft? and ents.contains?(edge) then
							findadjacent.call(edge, vertex)
						end
					end
				end
			end
		end
		
		findadjacent.call(ents[0], nil)
		
		faces = []
		curves = []
		removableedges = []
		curvepoints = [curvepoints]
		edges.each do |edge|
			if edge.faces.length > 0 then
				edge.faces.each do |face|
					unless faces.index(face)
						faces << face
						face.edges.each do |edge2|
							if edge2.curve and !edge2.soft? && !curves.index(edge2.curve) then
								curves << edge2.curve
								points = []
								edge2.curve.vertices.each do |vertex|
									points << vertex.position
								end
								curvepoints << points
								
								edge2.curve.edges.each do |edge3|
									unless edges.index(edge3)
										edges << edge3
									end
								end
							end
						end
					end
				end
			else
				removableedges << edge
			end
		end
		
		group = Sketchup.active_model.active_entities.add_group(faces)
		
		Sketchup.active_model.active_entities.erase_entities(removableedges)
		
		curveedges = []
		curvepoints.each do |points|
			curveedges << Sketchup.active_model.active_entities.add_curve(points)
		end
		Sketchup.active_model.selection.clear
		Sketchup.active_model.selection.add curveedges[0]
		
		if group then
			group.explode
		end
		
		Sketchup.active_model.commit_operation
		
		return edges
	end
	
	def Recurve.can_recurve_selection?()
		Sketchup.active_model and can_recurve?(Sketchup.active_model.selection)
	end
	
	def Recurve.recurve_selection()
		recurve Sketchup.active_model.selection
	end
	
	def Recurve.select_only(enttype, &filter)
		keeps = []
		Sketchup.active_model.selection.each do |ent|
			if ent.is_a? enttype then
				keeps << ent if !filter or filter.call(ent)
			end
		end
		
		Sketchup.active_model.selection.clear
		Sketchup.active_model.selection.add(keeps)
	end
end

# Add menu items
if( not file_loaded? "recurve.rb" )

	# Add a choice to the context menu
	UI.add_context_menu_handler do |menu|
		menu.add_separator
		
		selectOnlyMenu = menu.add_submenu("Select Only")
		selectOnlyMenu.add_item("Edges") { Recurve::select_only(Sketchup::Edge) }
		selectOnlyMenu.add_item("Hard Edges") { Recurve::select_only(Sketchup::Edge) { |edge| !edge.soft? } }
		selectOnlyMenu.add_item("Soft Edges") { Recurve::select_only(Sketchup::Edge) { |edge| edge.soft? } }
		selectOnlyMenu.add_item("Straight Edges") { Recurve::select_only(Sketchup::Edge) { |edge| !edge.curve } }
		selectOnlyMenu.add_item("Curved Edges") { Recurve::select_only(Sketchup::Edge) { |edge| edge.curve } }
		selectOnlyMenu.add_separator
		selectOnlyMenu.add_item("Faces") { Recurve::select_only(Sketchup::Face) }
		selectOnlyMenu.add_item("Convex Faces") { Recurve::select_only(Sketchup::Face) { |face| face.outer_loop.convex? } }
		selectOnlyMenu.add_item("Concave Faces") { Recurve::select_only(Sketchup::Face) { |face| !face.outer_loop.convex? } }
		selectOnlyMenu.add_item("Simple Faces") { Recurve::select_only(Sketchup::Face) { |face| face.loops.length == 1 } }
		selectOnlyMenu.add_item("Complex Faces") { Recurve::select_only(Sketchup::Face) { |face| face.loops.length != 1 } }
		selectOnlyMenu.add_separator
		selectOnlyMenu.add_item("Groups") { Recurve::select_only(Sketchup::Group) }
		
		valid = Recurve::can_select_curve_selection?
		if valid
			menu.add_item("Select Curve") { Recurve::select_curve_selection false }
			menu.add_item("Select Curve (Planar)") { Recurve::select_curve_selection true }
		end
		
		valid = Recurve::can_recurve_selection?
		if valid
			menu.add_item("Recurve Edges") { Recurve::recurve_selection }
		end
	end

	file_loaded "recurve.rb"
end

